// Copyright (C) 2009-2015, Software Development Laboratories, "Fish" (David B. Trout)
//////////////////////////////////////////////////////////////////////////////////////////
// TrayMe.cpp : Defines the entry point for the application.
//////////////////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "product.h"
#include "version.h"
#include "resource.h"

//////////////////////////////////////////////////////////////////////////////////////////
// Private window message id number values for our Tray Icon messages...

#define  TRAYICONID        12345        // Just an arbitrary number unique to us
#define  WM_MYTRAYICONMSG  WM_APP       // Message Id of our tray icon messages
#define  WM_MYABOUTCMD     WM_APP + 1   // Command Id of context menu "About" command
#define  WM_MYENABLECMD    WM_APP + 2   // Command Id of context menu "Enable/Disable" command
#define  WM_MYEXITCMD      WM_APP + 3   // Command Id of context menu "Exit" command

//////////////////////////////////////////////////////////////////////////////////////////
// Private global Variables...

HANDLE           g_hMainApp  = NULL;    // (our main application handle)
HMODULE          g_hHookDLL  = NULL;    // (handle of our hooking DLL)
HWND             g_hWnd      = NULL;    // (main application window handle)
HOOKUNHOOKFUNC*  g_pfAttach  = NULL;    // (ptr to the DLL's hooking function)
HOOKUNHOOKFUNC*  g_pfDetach  = NULL;    // (ptr to the DLL's unhooking function)
BOOL             g_bEnabled  = FALSE;   // (flag indicating our hooks are engaged)
HWND             g_hWndAbout = NULL;    // (About dialog window handle if being shown)
CRITICAL_SECTION g_Lock      = {0};     // (general purpose lock)

#define AcquireLock()       EnterCriticalSection( &g_Lock )
#define ReleaseLock()       LeaveCriticalSection( &g_Lock )

//////////////////////////////////////////////////////////////////////////////////////////
// Forward declarations for our private helper functions...

BOOL     InitInstance();
void     AddIconToSystemTray();
void     ExitInstance();
void     RemoveIconFromSystemTray();
void     ShowTrayContextMenu ( HWND hWnd );
INT_PTR  CALLBACK  MainProc  ( HWND, UINT, WPARAM, LPARAM );
INT_PTR  CALLBACK  AboutProc ( HWND, UINT, WPARAM, LPARAM );

//////////////////////////////////////////////////////////////////////////////////////////
// The main Windows Operating System entry-point for our application...

int  APIENTRY  _tWinMain
(
    HINSTANCE  hInstance,
    HINSTANCE  hPrevInstance,
    LPTSTR     pszCmdLine,
    int        nCmdShow
)
{
    // Save our instance handle since we'll likely need it...

    g_hMainApp = hInstance;

    // Perform application initialization...

    if (!InitInstance())
    {
        TRACE(_T("%s %4d: _tWinMain: InitInstance FAILED!\n"), _T(APPNAME), GetCurrentProcessId() );
        ExitInstance();
        return FALSE;
    }

    // Keep pumping windows messages until we're told to quit (or an error occurs)...

    MSG  msg;
    BOOL bRet;

    while ((bRet = GetMessage( &msg, NULL, 0, 0 )) != 0)
    {
        if (bRet < 0)
        {
            TRACE(_T("%s %4d: _tWinMain: GetMessage() ERROR! bRet = %d; exiting...\n"),
                _T(APPNAME), GetCurrentProcessId(), bRet );
            break;
        }

        TranslateMessage ( &msg );
        DispatchMessage  ( &msg );
    }

    // WM_QUIT was posted (or an error occurred). Exit the application...

    ExitInstance();

    return (int) msg.wParam;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Main application initialization...

BOOL  InitInstance ()
{
    InitializeCriticalSection( &g_Lock );

    // Load hooking DLL and retrieve needed entry-points...

    if (!(g_hHookDLL = LoadLibrary( _T(DLLNAME) )))
    {
        ErrorMessageBox( NULL, _T("LoadLibrary( ") _T(DLLNAME) _T(" ) failed!"),
            _T(APPNAME), MB_ICONWARNING | MB_OK );
        return FALSE;
    }

    if (0
        || !(g_pfAttach = (HOOKUNHOOKFUNC*) GetProcAddress( g_hHookDLL, HOOKFUNCNAME   ))
        || !(g_pfDetach = (HOOKUNHOOKFUNC*) GetProcAddress( g_hHookDLL, UNHOOKFUNCNAME ))
    )
    {
        ErrorMessageBox( NULL, _T("GetProcAddress for ") _T(DLLNAME) _T(" failed!"),
            _T(APPNAME), MB_ICONWARNING | MB_OK );
        VERIFY( FreeLibrary( g_hHookDLL ) && !(g_hHookDLL = NULL) );
        return FALSE;
    }

    // Create our main application dialog window (which is never actually shown)...

    g_hWnd = CreateDialog( (HINSTANCE) g_hMainApp, MAKEINTRESOURCE( IDD_MAINDIALOG ), NULL, MainProc );

    if (!g_hWnd)
    {
        ErrorMessageBox( NULL, _T("InitInstance: CreateDialog() failed!"),
            _T(APPNAME), MB_ICONWARNING | MB_OK );
        VERIFY( FreeLibrary( g_hHookDLL ) && !(g_hHookDLL = NULL) );
        return FALSE;
    }

    // (Note: ShowWindow never called so our main dialog is never shown)

    // Add ourselves to the system tray...

    AddIconToSystemTray();

    // Engage our hooks...

    VERIFY(g_bEnabled = g_pfAttach( g_hWnd, GetCurrentProcessId() ));
    return g_bEnabled;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Main application shutdown...

void  ExitInstance ()
{
    // Remove our hooks...

    TRACE(_T("%s %4d: ExitInstance: Removing hooks...\n"), _T(APPNAME), GetCurrentProcessId() );

    if (g_bEnabled)
        VERIFY( !(g_bEnabled = !g_pfDetach( g_hWnd, GetCurrentProcessId() )) );

    // Free the Hook DLL...

    TRACE(_T("%s %4d: ExitInstance: Freeing hook DLL...\n"), _T(APPNAME), GetCurrentProcessId() );

    if (g_hHookDLL)
        VERIFY( FreeLibrary( g_hHookDLL ) && !(g_hHookDLL = NULL) );

    // Remove ourselves from the system tray...

    RemoveIconFromSystemTray();

    DeleteCriticalSection( &g_Lock );
}

//////////////////////////////////////////////////////////////////////////////////////////
// Private message handling procedure for our main application dialog...

INT_PTR  CALLBACK  MainProc ( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    INT_PTR bMessageHandled = FALSE;     // (default)

    switch (uMsg)
    {
        case WM_INITDIALOG:
        {
            // Our main application dialog is never shown and if it ever is
            // it doesn't have any controls on it anyway so there's nothing
            // we need to do to initialize it...

            bMessageHandled = TRUE;
        }
        break;

        case WM_MYTRAYICONMSG:
        {
            // The user is interacting with our tray icon...

            bMessageHandled = TRUE;

            switch (lParam)
            {
                case WM_LBUTTONDBLCLK:
                {
                    // Double-clicking our tray icon displays our 'About' dialog...

                    DialogBox( (HINSTANCE) g_hMainApp, MAKEINTRESOURCE( IDD_ABOUTBOX ), hWnd, AboutProc );
                }
                break;

                case WM_RBUTTONDOWN:
                case WM_CONTEXTMENU:
                {
                    // Right-clicking our tray icon displays our context menu...

                    ShowTrayContextMenu( hWnd );
                }
                break;

                default:
                {
                    // (just ignore any/all other unknown/unexpected tray messages)
                }
                break;
            }
        }
        break; // case WM_MYTRAYICONMSG

        case WM_COMMAND:
        {
            // They selected a menu command from our context menu...

            bMessageHandled = TRUE;

            WORD wCommand = LOWORD( wParam );

            switch (wCommand)
            {
                case WM_MYABOUTCMD:
                {
                    // Only display the About dialog if it's not already being
                    // shown. If it is already being shown, ignore the request
                    // and switch focus to it instead.

                    AcquireLock();
                    {
                        if (g_hWndAbout)
                            SetFocus( g_hWndAbout );
                        else
                            DialogBox
                            (
                                (HINSTANCE) g_hMainApp,
                                MAKEINTRESOURCE( IDD_ABOUTBOX ),
                                hWnd,
                                AboutProc
                            );
                    }
                    ReleaseLock();
                }
                break; // case WM_MYABOUTCMD:

                case WM_MYENABLECMD:
                {
                    // They want to disable/enable our functionality...

                    if (g_bEnabled)
                    {
                        if (g_pfDetach( hWnd, GetCurrentProcessId() ))
                            g_bEnabled = FALSE;
                    }
                    else
                    {
                        if (g_pfAttach( hWnd, GetCurrentProcessId() ))
                            g_bEnabled = TRUE;
                        else
                            SafeMessageBox( hWnd, _T("Enable failed!"), _T(APPNAME), MB_ICONWARNING | MB_OK );
                    }
                }
                break; // case WM_MYENABLECMD:

                case WM_MYEXITCMD:
                {
                    // They want us to exit. Destroy our main dialog...

                    DestroyWindow( hWnd );
                }
                break; // case WM_MYEXITCMD

                default:
                {
                    // (just ignore any/all other unknown/unexpected menu commands)
                }
                break;

            } // end switch (wCommand)
        }
        break; // case WM_COMMAND

        case WM_CLOSE:
        {
            // Our main application is being asked to exit... They either
            // somehow pressed the Escape key or clicked the [X] Close box
            // on our dialog (which is unlikely since we never SHOW our
            // dialog!) or else something happened that caused the system
            // to ask us to 'close' our main application window...

            bMessageHandled = DestroyWindow( hWnd );
        }
        break;

        case WM_DESTROY:
        {
            // Our dialog is being destroyed/closed... (probably
            // because they selected "Exit" from our context menu).
            // Force our main _tWinMain message loop to end...

            PostQuitMessage( 0 );       // (time to exit!)
        }
        break; // case WM_DESTROY:

        // All other window messages we let the system do default processing for...

        default:
        {
            bMessageHandled = FALSE;
        }
        break;

    } // end switch (uMsg)

    return bMessageHandled;
}

//////////////////////////////////////////////////////////////////////////////////////////
// They right-clicked on our tray icon... Show them our context menu...

void  ShowTrayContextMenu ( HWND hWnd )
{
    // Retrieve the mouse position... This is where the menu will be shown...

    POINT pt; VERIFY( GetCursorPos( &pt ) );

    // Create a context menu...

    HMENU hMenu = CreatePopupMenu();

    if (hMenu)
    {
        // Insert our commands into our context menu...

        UINT uEnabledFlags = MF_BYPOSITION | (g_bEnabled ? MF_CHECKED : MF_UNCHECKED);

        VERIFY( InsertMenu( hMenu, -1, MF_BYPOSITION, WM_MYABOUTCMD,  _T("About")  ) );
        VERIFY( InsertMenu( hMenu, -1, MF_SEPARATOR,       0,             NULL     ) );
        VERIFY( InsertMenu( hMenu, -1, uEnabledFlags, WM_MYENABLECMD, _T("Enable") ) );
        VERIFY( InsertMenu( hMenu, -1, MF_SEPARATOR,       0,             NULL     ) );
        VERIFY( InsertMenu( hMenu, -1, MF_BYPOSITION, WM_MYEXITCMD,   _T("Exit")   ) );

        VERIFY( SetMenuDefaultItem( hMenu, WM_MYABOUTCMD, FALSE ) );

        // Note: you must bring the window to the foreground
        // or else the menu won't disappear when it should...

        VERIFY( SetForegroundWindow( hWnd ) );
        VERIFY( TrackPopupMenu( hMenu, TPM_BOTTOMALIGN, pt.x, pt.y, 0, hWnd, NULL ) );

        // Note: to prevent the menu from immediately disappearing the second
        // time it is displayed, you must force a task switch to the application
        // that called TrackPopupMenu by posting a benign message to it...

        VERIFY( PostMessage( hWnd, WM_NULL, 0, 0 ) );

        // Mission accomplished! Destroy our context menu
        // since we don't need it anymore...

        VERIFY( DestroyMenu( hMenu ) );
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
// Private message handling procedure for our 'About' dialog box...

INT_PTR  CALLBACK  AboutProc ( HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    INT_PTR bMessageHandled = FALSE;     // (default)

    switch (uMsg)
    {
        case WM_INITDIALOG:
        {
            // PROGRAMMING NOTE: our "MainProc" procedure has already
            // acquired the lock. All we need to do is release it.

            g_hWndAbout = hDlg;     // (indicate dialog being shown)
            ReleaseLock();          // (see above PROGRAMMING NOTE)

            // Initialize our 'About' dialog controls...

            bMessageHandled = TRUE;

            VERIFY( SetWindowText( hDlg, _T("About ") _T(APPNAME) ) );

            HWND hWnd;

            hWnd = GetDlgItem( hDlg, IDC_DESC_STATIC );
            VERIFY( SetWindowText( hWnd, _T(APPNAME) _T(" - ") PRODUCT ) );

            hWnd = GetDlgItem( hDlg, IDC_VERSION_STATIC );
            VERIFY( SetWindowText( hWnd, _T("Version: ") VERSION_STR ) );

            hWnd = GetDlgItem( hDlg, IDC_COPYRIGHT_STATIC );
            VERIFY( SetWindowText( hWnd, COPYRIGHT ) );

            hWnd = GetDlgItem( hDlg, IDC_COMPANY_STATIC );
            VERIFY( SetWindowText( hWnd, COMPANY ) );

            SetFocus( hDlg );
        }
        break;

        case WM_COMMAND:
        {
            // They clicked a button...  (or pressed the escape key)

            if (0
                || IDOK     == LOWORD( wParam )
                || IDCANCEL == LOWORD( wParam )
            )
            {
                // They clicked the 'OK' button or the [X] Close box
                // or pressed the Escape key to dismiss our dialog...

                // PROGRAMMING NOTE: our "MainProc" procedure will
                // release the lock once control is returned back
                // to it again by the EndDialog function call.

                AcquireLock();          // (see above PROGRAMMING NOTE)
                g_hWndAbout = NULL;     // (indicate dialog NOT being shown)

                bMessageHandled = TRUE;
                EndDialog( hDlg, LOWORD( wParam ) );
            }
        }
        break;

        // All other window messages we let the system do default processing for...

        default:
        {
            bMessageHandled = FALSE;
        }
        break;
    }

    return bMessageHandled;
}

//////////////////////////////////////////////////////////////////////////////////////////

void AddIconToSystemTray()
{
    TRACE(_T("%s %4d: AddIconToSystemTray\n"), _T(APPNAME), GetCurrentProcessId() );

    NOTIFYICONDATA ni; memset( &ni, 0, sizeof(ni) ); ni.cbSize = sizeof(ni);

    _tcscpy_s( ni.szTip, _countof( ni.szTip ), _T(APPNAME) _T(" - ") PRODUCT );

    ni.uFlags = NIF_ICON | NIF_MESSAGE | NIF_TIP;
    ni.hWnd   = g_hWnd;
    ni.uID    = TRAYICONID;
    ni.hIcon  = (HICON) LoadImage
    (
        (HINSTANCE) g_hMainApp,
        MAKEINTRESOURCE( IDI_ICON ),
        IMAGE_ICON,
        GetSystemMetrics( SM_CXSMICON ),
        GetSystemMetrics( SM_CYSMICON ),
        LR_DEFAULTCOLOR
    );
    ni.uCallbackMessage = WM_MYTRAYICONMSG;

    VERIFY( Shell_NotifyIcon( NIM_ADD, &ni ) );

    if (ni.hIcon)
        DestroyIcon( ni.hIcon );   // (no longer needed)
}

//////////////////////////////////////////////////////////////////////////////////////////

void RemoveIconFromSystemTray()
{
    TRACE(_T("%s %4d: RemoveIconFromSystemTray\n"), _T(APPNAME), GetCurrentProcessId() );

    NOTIFYICONDATA ni; memset( &ni, 0, sizeof(ni) ); ni.cbSize = sizeof(ni);

    ni.hWnd = g_hWnd;
    ni.uID  = TRAYICONID;

    VERIFY( Shell_NotifyIcon( NIM_DELETE, &ni ) );
}

//////////////////////////////////////////////////////////////////////////////////////////
