// Copyright (C) 2009-2012, Software Development Laboratories, "Fish" (David B. Trout)
//////////////////////////////////////////////////////////////////////////////////////////
// TrayMeDLL.cpp : Defines the entry point for the DLL application.
//////////////////////////////////////////////////////////////////////////////////////////
//
//  Change History:
//
//  08/07/09    1.0.0   Fish    Created.
//  03/01/10    1.0.1   Fish    Declare g_nTrayCount and g_bEnabled as volatile since
//                              they're both updated by multiple different processes.
//
//////////////////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "product.h"
#include "version.h"

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Private application global variables NOT shared by all applications...

static HWND       g_hWnd         = NULL;    // Current app's main window handle
static HINSTANCE  g_hHookDLL     = NULL;    // This DLL's module instance handle
static bool       g_bHidden      = false;   // true if we're minimized to tray icon

//////////////////////////////////////////////////////////////////////////////////////////
// Function forward references...

void   MainWndHook( WPARAM wParam, LPARAM lParam );   // (main window hook logic)
void   SysCmdHook ( WPARAM wParam, LPARAM lParam );   // (system menu hook logic)

LRESULT  CALLBACK  SysCmdHookProc  ( int nCode, WPARAM wParam, LPARAM lParam );
LRESULT  CALLBACK  MainWndHookProc ( int nCode, WPARAM wParam, LPARAM lParam );

void   MinimizeToSystemTray  ();        // Minimizes application to system tray
void   AddIconToSysTray      ();        // Add app icon to system tray ONLY
void   RestoreFromSystemTray ();        // Restores application from system tray
HICON  GetSmallIconHandle    ();        // Retrieve application's small icon

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Public global variables in shared data segment shared by ALL hooked processes...

#pragma  data_seg( push, save, ".share" )

//    PROGRAMMING NOTE: the following global variables are used (referenced)
//    by multiple different processes but do not need to be declared volatile
//    since they are never ever modified once they have been initialized.

DWORD   g_dwMainProcId    = 0;          // Process Id of our main application
UINT    g_nTRAYICONMSG    = 0;          // System tray icon message
UINT    g_nUNHIDEMSG      = 0;          // Request to unhide from tray
UINT    g_nTASKBARCREATED = 0;          // WM_TASKBARCREATED message
HHOOK   g_hMainWndHook    = NULL;       // Primary window message hook proc
HHOOK   g_hSysCmdHook     = NULL;       // System menu command hook procecdure

#define  ISHOOKED()  ( g_hMainWndHook && g_hSysCmdHook )

//    PROGRAMMING NOTE: the following variables must be declared volatile
//    since they are updated externally by another process than their own.

volatile BOOL g_bEnabled   = FALSE;     // TRUE to allow going to system tray
volatile LONG g_nTrayCount = 0;         // Count of apps placed in system tray

#pragma  data_seg( pop, save )          // (back to private UNshared data seg)

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// Our main Tray Icon application is asking us to install our hooks...

BOOL  WINAPI  DoHook ( HWND hWnd, DWORD dwProcessId )
{
    // NOTE: This function should only ever called by our main TrayMe application.

    if (0
        || g_bEnabled
        || ISHOOKED()
    )
    {
        TRACE(_T("%s %4d: DoHook: Already hooked!\n"), _T(APPNAME), GetCurrentProcessId() );
        ASSERT(FALSE);
        return FALSE;
    }

    TRACE(_T("%s %4d: DoHook: Installing hooks...\n"), _T(APPNAME), GetCurrentProcessId() );

    // Save these so we can detect/prevent "rogue" callers...

    g_hWnd         = hWnd;
    g_dwMainProcId = dwProcessId;

    // Create needed message ids...

    VERIFY( g_nTRAYICONMSG    = RegisterWindowMessage( _T(TRAYICONMSGNAME) ) );
    VERIFY( g_nUNHIDEMSG      = RegisterWindowMessage( _T(UNHIDEMSGNAME)   ) );
    VERIFY( g_nTASKBARCREATED = RegisterWindowMessage( _T(TASKBARCREATED)  ) );

    // Hook all running processes...

    DWORD dwLastError = 0;

    if (!(g_hMainWndHook = SetWindowsHookEx( WH_CALLWNDPROC, MainWndHookProc, g_hHookDLL, 0 )))
    {
        dwLastError = GetLastError();
        TRACE(_T("%s %4d: DoHook: SetWindowsHookEx( WH_CALLWNDPROC ) FAILED! rc=%d (0x%08.8X)\n"),
            _T(APPNAME), GetCurrentProcessId(), dwLastError, dwLastError );
    }

    if (!(g_hSysCmdHook = SetWindowsHookEx( WH_GETMESSAGE,  SysCmdHookProc,  g_hHookDLL, 0 )))
    {
        dwLastError = GetLastError();
        TRACE(_T("%s %4d: DoHook: SetWindowsHookEx( WH_GETMESSAGE ) FAILED! rc=%d (0x%08.8X)\n"),
            _T(APPNAME), GetCurrentProcessId(), dwLastError, dwLastError );
    }

    return (g_bEnabled = ISHOOKED());
}

//////////////////////////////////////////////////////////////////////////////////////////
// Our main Tray Icon application is asking us to uninstall our hooks...

BOOL  WINAPI  DoUnHook ( HWND hWnd, DWORD dwProcessId )
{
    // NOTE: This function should only ever called by our main TrayMe application.

    if (0
        || !g_bEnabled
        || !ISHOOKED()
        || !g_dwMainProcId
        ||  g_dwMainProcId != GetCurrentProcessId()
    )
    {
        TRACE(_T("%s %4d: DoUnHook: Already unhooked! (or not original hooking process)\n"),
            _T(APPNAME), GetCurrentProcessId() );

        ASSERT(FALSE);   // (should only be called by our main app!)
        return FALSE;
    }

    ASSERT( g_hWnd && hWnd == g_hWnd );

    // Clear the "Enabled" flag right away to stop everyone
    // from even TRYING to send themselves to the system tray...

    g_bEnabled = FALSE;

    // Ask everyone already in the tray to unhide themselves...

    if (g_nTrayCount)
    {
        TRACE(_T("%s %4d: DoUnHook: g_nTrayCount=%d; Broadcasting UNHIDEMSG\n"),
            _T(APPNAME), GetCurrentProcessId(), g_nTrayCount );

        do
        {
            // PROGRAMMING NOTE: the following MUST be a SendMessage
            // and NOT a PostMessage! Otherwise "TrayMe!"ed programs
            // NEVER automatically unhide themselves, and the TrayMe
            // program NEVER exits until ALL "TrayMe"ed applications
            // are eventually MANUALLY unhidden! (*VERY* weird!)

            SendMessage( HWND_BROADCAST, g_nUNHIDEMSG, 0, 0 );

            // (wait for everyone to remove themselves from the tray)

            for (int i = UNHOOKTIMEOUT; g_nTrayCount && i > 0; i -= UNHOOKSLEEP)
                Sleep( UNHOOKSLEEP );

            if (!g_nTrayCount) break;  // (expected outcome)

            // (ask what to do if some apps still remain in the tray)

            TRACE(_T("%s %4d: DoUnHook: g_nTrayCount=%d; Not everyone has unhidden themselves!\n"),
                _T(APPNAME), GetCurrentProcessId(), g_nTrayCount );

            int rc = SafeMessageBox( g_hWnd,
                _T("Not everyone has unhidden themselves!"),
                _T(" Unhook Error!"), MB_ICONWARNING | MB_ABORTRETRYIGNORE
            );

            if (IDABORT  == rc) {g_bEnabled = ISHOOKED(); return FALSE;}
            if (IDRETRY  == rc) continue;
            if (IDIGNORE == rc) break;
        }
        while (g_nTrayCount);
    }

    // Unhook all running processes...

    TRACE(_T("%s %4d: DoUnHook: calling UnhookWindowsHookEx\n"),
        _T(APPNAME), GetCurrentProcessId());

    if (UnhookWindowsHookEx( g_hMainWndHook )) g_hMainWndHook = NULL;
    if (UnhookWindowsHookEx( g_hSysCmdHook  )) g_hSysCmdHook  = NULL;

    // Finally, post a benign message to all top-level windows
    // in order to force them to disengage from our Hook DLL...

    // PROGRAMMING NOTE: the following may be a SendMessage -OR-
    // a PostMessage, it really doesn't matter, but a SendMessage
    // seems safer since the WM_NULL message really is benign and
    // we do want to ensure that all TrayMe users have disengaged.

    TRACE(_T("%s %4d: DoUnHook: Broadcasting WM_NULL\n"),
        _T(APPNAME), GetCurrentProcessId());

    SendMessage( HWND_BROADCAST, WM_NULL, 0, 0 );

    return !(g_bEnabled = ISHOOKED());
}

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
// All remaining functions in this module are shared by ALL hooked processes...

#pragma  comment( linker, "/section:.share,RWS" )

// The below two hook procedure callback functions can be called by any process
// at any time, including by our main TrayMe application itself. If it's called
// by some other process then we "do our thing" (perform our hook processing),
// but if it's ever called by our own main TrayMe application then we do nothing
// since our main TrayMe application's main window is never shown and thus never
// needs to be "TrayMe"ed... (i.e. we only send OTHER processes to the system
// tray, but we have absolutely no need to ever send OURSELVES to the tray since
// we should ALWAYS already be THERE!)...

LRESULT  CALLBACK  MainWndHookProc ( int nCode, WPARAM wParam, LPARAM lParam )
{
    if (1
        && nCode >= 0
        && g_dwMainProcId
        && g_dwMainProcId != GetCurrentProcessId()
    )
        MainWndHook( wParam, lParam );  // (perform hook processing)

    return CallNextHookEx( g_hMainWndHook, nCode, wParam, lParam );
}

LRESULT  CALLBACK  SysCmdHookProc ( int nCode, WPARAM wParam, LPARAM lParam )
{
    if (1
        && nCode >= 0
        && g_dwMainProcId
        && g_dwMainProcId != GetCurrentProcessId()
    )
        SysCmdHook( wParam, lParam );   // (perform hook processing)

    return CallNextHookEx( g_hSysCmdHook, nCode, wParam, lParam );
}

//////////////////////////////////////////////////////////////////////////////////////////
// Our special main window hook processing logic...

void  MainWndHook ( WPARAM wParam, LPARAM lParam )
{
    CWPSTRUCT* pHookMsg = (CWPSTRUCT*) lParam;

    if (!g_bHidden)
    {
        // They're currently NOT hidden to the system tray so the only
        // messages we're interested in are whether the System Menu is
        // being shown or being closed so we can either add or remove
        // our special "TrayMe!" command to/from their system menu...

        HMENU hSysMenu = GetSystemMenu( pHookMsg->hwnd, FALSE );

        // Are they about to display the system menu?

        if (g_bEnabled && pHookMsg->message == WM_INITMENUPOPUP)
        {
            BOOL bSysMenu = HIWORD( pHookMsg->lParam ) ? TRUE : FALSE;

            if (bSysMenu && IsMenu( hSysMenu ))
            {
                // Append our "Tray Me!" command to their system menu...

                TRACE(_T("%s %4d: MainWndHook: AppendMenu\n"), _T(APPNAME), GetCurrentProcessId());

                g_hWnd = pHookMsg->hwnd;  // (save for later RemoveMenu)
                VERIFY( AppendMenu( hSysMenu, MF_STRING, g_nTRAYICONMSG, _T(TRAYME) ) );
            }
        }

        // Are they closing their system menu? (or any other menu)

        // "If the high-order word of wParam contains 0xFFFF
        //  and the lParam parameter contains NULL, then the
        //  system has closed the menu."

        else if (1
            &&         pHookMsg->message  == WM_MENUSELECT
            &&         pHookMsg->hwnd     == g_hWnd
            && HIWORD( pHookMsg->wParam ) == 0xFFFF   // (see above)
            &&         pHookMsg->lParam   == NULL     // (see above)
        )
        {
            // PROGRAMMING NOTE: since we can't tell whether this is
            // the system menu being closed or some other menu being
            // closed, we ignore any error that might occur from the
            // call to RemoveMenu since we may have already removed
            // ourselves from the system menu already. Also note we
            // cannot simply call GetSystemMenu with bRevert = TRUE
            // since they may have inserted their own commands into
            // the system menu so we cannot arbitrarily remove them!

            TRACE(_T("%s %4d: MainWndHook: RemoveMenu\n"), _T(APPNAME), GetCurrentProcessId());

            RemoveMenu( hSysMenu, g_nTRAYICONMSG, MF_BYCOMMAND );
        }
    }

    // If they're currently minimized to the system tray and they've
    // either double-clicked on their tray icon or they're exiting,
    // or if this a request by our main TrayMe application to force
    // them to show themself, then remove them from the system tray...

    else if (g_bHidden &&
    (0
        // Double-click on system tray icon?

        || (1
            && pHookMsg->message == g_nTRAYICONMSG
            && pHookMsg->lParam  == WM_LBUTTONDBLCLK
            && pHookMsg->hwnd    == g_hWnd
           )

        // Main window exiting?

        || (1
            && (0
                || pHookMsg->message == WM_CLOSE
                || pHookMsg->message == WM_DESTROY
               )
            &&     pHookMsg->hwnd    == g_hWnd
           )

        // Forced unhide request from main app?

        || (pHookMsg->message == g_nUNHIDEMSG)
    ))
    {
        TRACE(_T("%s %4d: MainWndHook: RestoreFromSystemTray\n"), _T(APPNAME), GetCurrentProcessId());

        RestoreFromSystemTray();  // (show app window again)
    }

    // We must remember to ALWAYS re-add their icon to the system tray
    // again (if they were minimized to it that is!) whenever Windows
    // Explorer ever restarts itself for any reason (e.g. after it has
    // crashed for example)...

    else if (g_bHidden && pHookMsg->message == g_nTASKBARCREATED)
    {
        TRACE(_T("%s %4d: MainWndHook: AddIconToSysTray\n"), _T(APPNAME), GetCurrentProcessId());

        AddIconToSysTray();   // (recreate their system tray icon)
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
// Our System Menu command hook processing logic...

void  SysCmdHook ( WPARAM wParam, LPARAM lParam )
{
    // Is this our special "Tray Me!" system command
    // that we inserted into their System Menu?

    MSG* pMSG = (MSG*) lParam;

    if (1
        &&  g_bEnabled    // (if we're enabled)
        &&  g_hWnd        // (we have a window)
        && !g_bHidden     // (and it's not hidden)

        &&         pMSG->message  == WM_SYSCOMMAND
        && LOWORD( pMSG->wParam ) == g_nTRAYICONMSG
        &&         pMSG->hwnd     == g_hWnd
    )
    {
        // Yes, add them to the system tray and
        // automatically hide their main window...

        TRACE(_T("%s %4d: SysCmdHook: MinimizeToSystemTray\n"), _T(APPNAME), GetCurrentProcessId());

        MinimizeToSystemTray();
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
// (Note: also hides the window)

void MinimizeToSystemTray()
{
    TRACE(_T("%s %4d: MinimizeToSystemTray\n"), _T(APPNAME), GetCurrentProcessId() );

    ASSERT( !g_bHidden && g_hWnd );
    AddIconToSysTray();

    // Now hide their window...

    ShowWindow( g_hWnd, SW_HIDE );
    g_bHidden = true;
    InterlockedIncrement( &g_nTrayCount );
}

//////////////////////////////////////////////////////////////////////////////////////////
// (only adds icon, does NOT hide window)  (used to recreate tray icon if lost)

void AddIconToSysTray()
{
    TRACE(_T("%s %4d: AddIconToSysTray\n"), _T(APPNAME), GetCurrentProcessId() );

    ASSERT( g_hWnd );

    // (use window's title for tray icon's tooltip)

    TCHAR szWindowTitle[ MAXTITLELEN ];
    GetWindowText( g_hWnd, szWindowTitle, _countof( szWindowTitle ) );

    NOTIFYICONDATA ni; memset( &ni, 0, sizeof(ni) ); ni.cbSize = sizeof(ni);
    _tcsncpy_s( ni.szTip, _countof( ni.szTip ), szWindowTitle, _TRUNCATE );

    ni.uFlags            =  NIF_ICON | NIF_MESSAGE | NIF_TIP;
    ni.hWnd              =  g_hWnd;
    ni.uID               =  g_nTRAYICONMSG;
    ni.uCallbackMessage  =  g_nTRAYICONMSG;
    ni.hIcon             =  GetSmallIconHandle();

    VERIFY( Shell_NotifyIcon( NIM_ADD, &ni ) );
    VERIFY( DestroyIcon( ni.hIcon ) );
}

//////////////////////////////////////////////////////////////////////////////////////////
// (Note: also unhides (shows) the window)

void RestoreFromSystemTray()
{
    TRACE(_T("%s %4d: RestoreFromSystemTray\n"), _T(APPNAME), GetCurrentProcessId() );

    ASSERT( g_bHidden && g_hWnd );

    NOTIFYICONDATA ni; memset( &ni, 0, sizeof(ni) ); ni.cbSize = sizeof(ni);

    ni.hWnd  =  g_hWnd;
    ni.uID   =  g_nTRAYICONMSG;

    VERIFY( Shell_NotifyIcon( NIM_DELETE, &ni ) );

    // Now show their window...

    ShowWindow( g_hWnd, SW_SHOW );
    g_bHidden = false;
    InterlockedDecrement( &g_nTrayCount );
}

//////////////////////////////////////////////////////////////////////////////////////////
// Helper function to retrieve an application's icon...

HICON GetSmallIconHandle()
{
    TRACE(_T("%s %4d: GetSmallIconHandle\n"), _T(APPNAME), GetCurrentProcessId() );

    // PROGRAMMING NOTE: the returned icon handle MUST be
    //                   one which can later be destroyed!

    HICON hIcon = NULL;

    // Try getting it directly from the application window if possible...
    // (Note: in my experience this RARELY works for some odd reason)

    if (g_hWnd)
        hIcon = (HICON) SendMessage( g_hWnd, WM_GETICON, ICON_SMALL2, 0 );

    if (hIcon)
    {
        TRACE(_T("%s %4d: GetSmallIconHandle: WM_GETICON worked!\n"), _T(APPNAME), GetCurrentProcessId() );
        hIcon = CopyIcon( hIcon );  // (so we can destroy it)
        if (hIcon)
            return hIcon;
    }

    // Try getting it directly from the application's executable file...

    TCHAR  szModuleFilename [ _MAX_PATH ];

    GetModuleFileName( NULL, szModuleFilename, _countof( szModuleFilename ) );

    UINT uFlags =
        0
        | SHGFI_ICON
        | SHGFI_USEFILEATTRIBUTES
        | SHGFI_SMALLICON
        ;

    SHFILEINFO sfi;
    SHGetFileInfo( szModuleFilename, FILE_ATTRIBUTE_NORMAL, &sfi, sizeof( sfi ), uFlags );
    hIcon = sfi.hIcon;

    // If we still don't have one, default to the system question-mark icon...

    if (!hIcon)
        hIcon = LoadIcon( NULL, IDI_QUESTION );

    return hIcon;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Main operating system DLL entry-point function...

BOOL  WINAPI  DllMain
(
    HANDLE  hModule,        // Handle of this DLL in context of calling process
    DWORD   dwReason,       // Reason we're being called (see docs)
    LPVOID  pReserved       // Reserved; must be NULL
)
{
    g_hHookDLL = (HINSTANCE) hModule;     // (needed to set our hooks)
    return TRUE;
}

//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
